@@ -25,23 +25,27 @@
 
 
 import numpy as np
-import dill
 from tqdm import tqdm
+import multiprocessing as mp
+from contextlib import closing
+import platform
 
-from sicor.Tools.EnMAP.LUT import read_lut_enmap_formatted, interpol_lut_c, get_data_file
+from py_tools_ds.processing.progress_mon import ProgressBar
+
+from sicor.Tools.EnMAP.LUT import read_lut_enmap_formatted, interpol_lut
 from sicor.Tools.EnMAP.conversion import generate_filter
 from sicor.Tools.EnMAP.metadata import varsol
+from sicor.Tools.EnMAP.multiprocessing import SharedNdarray, initializer, mp_progress_bar
 
 
-def wv_band_ratio(data, water_msk, land_only, fn_table, vza, sza, dem, aot, raa, intp_wvl, intp_fwhm, jday, month, idx,
-                  disable=False):
+def wv_band_ratio(data, water_msk, fn_table, sol_model, vza, sza, dem, aot, raa, intp_wvl, intp_fwhm, jday, month, idx,
+                  processes, disable=False):
     """Band ratio water vapor retrieval.
 
     :param data:      image dataset
     :param water_msk: water mask
-    :param land_only: if True, CWV first guess is calculated for land surfaces only;
-                      if False, all image pixels (land + water) are processed
     :param fn_table:  path to radiative transfer LUT
+    :param sol_model: dictionary containing solar irradiance model ('wvl', 'sol_irr')
     :param vza:       viewing zenitg angle
     :param sza:       sun zenith angle
     :param dem:       digital elevation model, same shape as data
@@ -53,6 +57,7 @@ def wv_band_ratio(data, water_msk, land_only, fn_table, vza, sza, dem, aot, raa,
     :param month:     acquisition month
     :param idx:       indices of instrument channels, which should be used for retrieval
                       (should be approx. 870, 900 and 940 nm)
+    :param processes: number of CPUs for multiprocessing
     :param disable:   if True, progressbar during retrieval is disabled; default: False
     :return:          water vapor image
     """
@@ -91,8 +96,8 @@ def wv_band_ratio(data, water_msk, land_only, fn_table, vza, sza, dem, aot, raa,
         for jj, cwv in enumerate(cwvs):
 
             vtest = np.asarray([vza, sza, hsf, aot, raa, cwv])
-            f_int = interpol_lut_c(lut1=lut1_res, lut2=lut2_res, xnodes=xnodes, nm_nodes=nm_nodes, ndim=ndim,
-                                   x_cell=x_cell, vtest=vtest, intp_wvl=intp_wvl)
+            f_int = interpol_lut(lut1=lut1_res, lut2=lut2_res, xnodes=xnodes, nm_nodes=nm_nodes, ndim=ndim,
+                                 x_cell=x_cell, vtest=vtest, intp_wvl=intp_wvl)
 
             f_int_l0 = f_int[0, :] * 1.e+3
             f_int_edir = f_int[1, :] * 1.e+3
@@ -105,10 +110,8 @@ def wv_band_ratio(data, water_msk, land_only, fn_table, vza, sza, dem, aot, raa,
                 l_toa = (f_int_l0 + f_int_ee * rho / np.pi / (1 - f_int_ss * rho)) * fac
                 l_toa_lut[ii, jj, kk, :] = l_toa
 
-    path_sol = get_data_file(module_name="sicor", file_basename="solar_irradiances_400_2500_1.dill")
-    with open(path_sol, "rb") as fl:
-        solar_lut = dill.load(fl)
-    solar_res = solar_lut @ s_norm
+    s_norm_sol = generate_filter(wvl_m=sol_model["wvl"], wvl=intp_wvl, wl_resol=intp_fwhm)
+    solar_res = sol_model["sol_irr"] @ s_norm_sol
 
     rfl_1_img = data[:, :, idx[0]] / solar_res[idx[0]]
     rfl_2_img = data[:, :, idx[1]] / solar_res[idx[1]]
@@ -146,40 +149,69 @@ def wv_band_ratio(data, water_msk, land_only, fn_table, vza, sza, dem, aot, raa,
 
     s0 = solar_res[idx[1]]
     ll_cal = np.pi * np.ndarray.flatten(data[:, :, idx[1]])[:] / (s0 * np.cos(np.deg2rad(sza)))
-    wv_arr = np.empty(cnt_land)
-    wv_arr[:] = np.nan
 
-    for ind in tqdm(range(0, cnt_land), disable=disable):
-        if land_only and water_msk_flat[ind] != 1:
-            pass
-        else:
-            dem_fac_pix = dem_fac[ind]
+    global _globs
+    _globs = dict(water_msk_flat=water_msk_flat, dem_fac=dem_fac, ll_cal=ll_cal, rhos=rhos, rfl_sl_img=rfl_sl_img,
+                  rfl_sl_gr=rfl_sl_gr, cf_arr=cf_arr, alog_rat_wv_img=alog_rat_wv_img)
 
-            if ll_cal[ind] <= rhos[0]:
-                ll_cal[ind] = rhos[0] + 0.01 * rhos[0]
-            ll_cal_low = ll_cal[ind] >= rhos
-            ll_cal_high = ll_cal[ind] <= rhos
-            idx_low = np.where(ll_cal_low)
-            idx_high = np.where(ll_cal_high)
+    _globs["__wv_arr__"] = SharedNdarray(dims=list((cnt_land, 1)))
 
-            rfl_fac = (ll_cal[ind] - rhos[idx_low[0][-1]]) / (rhos[idx_high[0][0]] - rhos[idx_low[0][-1]])
-            rfl_fac_pix = rfl_fac
+    # check if operating system is 'Windows'; in that case, multiprocessing is currently not working
+    # TODO: enable Windows compatibility for multiprocessing
+    if platform.system() == "Windows" or processes == 1:
+        initializer(globals(), _globs)
+        for ind in tqdm(range(0, cnt_land), disable=disable):
+            _compute_wv(ind)
+    else:
+        with closing(mp.get_context("fork").Pool(processes=processes, initializer=initializer,
+                                                 initargs=(globals(), _globs))) as pool:
+            results = pool.map_async(_compute_wv, range(0, cnt_land), chunksize=1)
+            if not disable:
+                bar = ProgressBar(prefix='\tprogress:')
+            while True:
+                if not disable:
+                    mp_progress_bar(iter_list=range(0, cnt_land), results=results, bar=bar)
+                if results.ready():
+                    results.get()
+                    break
 
-            rfl_sl = np.ndarray.flatten(rfl_sl_img)[ind]
-            rfl_sl_pix = (rfl_sl - rfl_sl_gr[0]) / (rfl_sl_gr[1] - rfl_sl_gr[0])
-
-            cf_int = (1 - dem_fac_pix) * (1 - rfl_fac_pix) * (1 - rfl_sl_pix) * cf_arr[:, 0, idx_low[0][-1], 0] + (
-                    1 - dem_fac_pix) * (1 - rfl_fac_pix) * rfl_sl_pix * cf_arr[:, 0, idx_low[0][-1], 1] + (
-                             1 - dem_fac_pix) * rfl_fac_pix * (1 - rfl_sl_pix) * cf_arr[:, 0, idx_high[0][0], 0] + (
-                             1 - dem_fac_pix) * rfl_fac_pix * rfl_sl_pix * cf_arr[:, 0, idx_high[0][0], 1] + dem_fac_pix * (
-                             1 - rfl_fac_pix) * (1 - rfl_sl_pix) * cf_arr[:, 1, idx_low[0][-1], 0] + dem_fac_pix * (
-                             1 - rfl_fac_pix) * rfl_sl_pix * cf_arr[:, 1, idx_low[0][-1], 1] + dem_fac_pix * rfl_fac_pix * (
-                             1 - rfl_sl_pix) * cf_arr[:, 1, idx_high[0][0],
-                                               0] + dem_fac_pix * rfl_fac_pix * rfl_sl_pix * cf_arr[:, 1, idx_high[0][0], 1]
-
-            wv = cf_int[2] + alog_rat_wv_img[ind] * cf_int[1] + alog_rat_wv_img[ind] * alog_rat_wv_img[ind] * cf_int[0]
-            wv_arr[ind] = wv
-
-    wv_arr_img = np.reshape(wv_arr, (data.shape[:2]))
+    wv_arr_img = np.reshape(_globs["__wv_arr__"].np, (data.shape[:2]))
 
     return wv_arr_img
+
+
+_globs = dict()
+
+
+# noinspection PyUnresolvedReferences
+def _compute_wv(ind):
+    if _globs["water_msk_flat"][ind] != 1:
+        pass
+    else:
+        dem_fac_pix = _globs["dem_fac"][ind]
+
+        if _globs["ll_cal"][ind] <= _globs["rhos"][0]:
+            _globs["ll_cal"][ind] = _globs["rhos"][0] + 0.01 * _globs["rhos"][0]
+        ll_cal_low = _globs["ll_cal"][ind] >= _globs["rhos"]
+        ll_cal_high = _globs["ll_cal"][ind] <= _globs["rhos"]
+        idx_low = np.where(ll_cal_low)
+        idx_high = np.where(ll_cal_high)
+
+        rfl_fac = (_globs["ll_cal"][ind] - _globs["rhos"][idx_low[0][-1]]) / (
+            _globs["rhos"][idx_high[0][0]] - _globs["rhos"][idx_low[0][-1]])
+        rfl_fac_pix = rfl_fac
+
+        rfl_sl = np.ndarray.flatten(_globs["rfl_sl_img"])[ind]
+        rfl_sl_pix = (rfl_sl - _globs["rfl_sl_gr"][0]) / (_globs["rfl_sl_gr"][1] - _globs["rfl_sl_gr"][0])
+
+        cf_int = (1 - dem_fac_pix) * (1 - rfl_fac_pix) * (1 - rfl_sl_pix) * _globs["cf_arr"][:, 0, idx_low[0][-1], 0] +\
+                 (1 - dem_fac_pix) * (1 - rfl_fac_pix) * rfl_sl_pix * _globs["cf_arr"][:, 0, idx_low[0][-1], 1] + \
+                 (1 - dem_fac_pix) * rfl_fac_pix * (1 - rfl_sl_pix) * _globs["cf_arr"][:, 0, idx_high[0][0], 0] + \
+                 (1 - dem_fac_pix) * rfl_fac_pix * rfl_sl_pix * _globs["cf_arr"][:, 0, idx_high[0][0], 1] + \
+                 dem_fac_pix * (1 - rfl_fac_pix) * (1 - rfl_sl_pix) * _globs["cf_arr"][:, 1, idx_low[0][-1], 0] + \
+                 dem_fac_pix * (1 - rfl_fac_pix) * rfl_sl_pix * _globs["cf_arr"][:, 1, idx_low[0][-1], 1] + \
+                 dem_fac_pix * rfl_fac_pix * (1 - rfl_sl_pix) * _globs["cf_arr"][:, 1, idx_high[0][0], 0] + \
+                 dem_fac_pix * rfl_fac_pix * rfl_sl_pix * _globs["cf_arr"][:, 1, idx_high[0][0], 1]
+
+        __wv_arr__[ind, :] = cf_int[2] + _globs["alog_rat_wv_img"][ind] * cf_int[1] + _globs["alog_rat_wv_img"][ind] * \
+                             _globs["alog_rat_wv_img"][ind] * cf_int[0]
